#ifndef LANDSCAPETOPOLOGY_HH_
#define LANDSCAPETOPOLOGY_HH_

#include "ell/State.hh"

#include <string>

// get information on available hashes or maps
#include "biu/HashMap.hh"

// include best available hash or map
#if HAVE_UNORDERED_MAP == 1
	#include <unordered_map>
#else
#if HAVE_TR1_UNORDERED_MAP == 1
	#include <tr1/unordered_map>
#else
#if HAVE_GNU_HASH_MAP == 1
	#include <ext/hash_map>
#else
	#include <map>
#endif
#endif
#endif


//##############################################################################
//##############################################################################

namespace ell {

/**
 * An abstract representation of the landscapes topology that represents minima 
 * and adjacency information like saddle points or barriers between them.
 * 
 * @author Hannes Kochniss, Martin Mann
 */
class LandscapeTopology {

public:
	

	// set typedef for best available hash or map
	#if HAVE_UNORDERED_MAP == 1
		typedef std::unordered_map< int, size_t > Int2SizeT_MAP;
	#else
	#if HAVE_TR1_UNORDERED_MAP == 1
		typedef std::tr1::unordered_map< int, size_t > Int2SizeT_MAP;
	#else
	#if HAVE_GNU_HASH_MAP == 1
		typedef __gnu_cxx::hash_map< int, size_t > Int2SizeT_MAP;
	#else
		typedef std::map< int, size_t > Int2SizeT_MAP;
	#endif
	#endif
	#endif

public:  /////////   STATIC CONSTANTS  ////////////////////////////////////////
	
	//! Index value that is used to propagate the information about invalid
	//! indexing.
	static const size_t INVALID_INDEX;
	
	  //! Avogadro constant N_A = 6.02214179 * 10^(23)
	  //! unit : 1/mol
	static const double AVOGADRO_CONSTANT_NA;
	  //! Gas constant R = 8.314472
	  //! unit : Joule / (Kelvin * mol)
	static const double GAS_CONSTANT_R;
	  //! Boltzmann constant K_b = R / N_A = 1.3806504 * 10^(-23) 
	  //! unit : Joule / Kelvin
	static const double BOLTZMANN_CONSTANT_KB;
	
	  //! Conversion factor * Joule --> kcal = 0.000239
	static const double JOULE_TO_KCAL;

public:  /////////   STATIC VARIABLES  ////////////////////////////////////////
	
	  //! The kT value used to calculate the Boltzmann weight e^(-E/kT) of a 
	  //! state in the landscape with an energy of E.
	  //! NOTE: Has to be set according to the used energy function !!!
	static double BOLTZMANN_KT;
	
	  //! The penalty used to score a missing minimum when comparing two
	  //! landscape topologies using the 'getDistance' method.
	  //! NOTE: Has to be set according to the used energy function and the
	  //! weighting of missing minima in the distance score. If missing minima
	  //! are of no relevance set this value to 0.0!
	static double MISSING_MINIMUM_PENALTY;
	
public:
	
	//! Destruction
	virtual 
	~LandscapeTopology();
	
	//! Sets a saddle point that connects two minima 
	//! @param i index of the first minimum to connect
	//! @param j index of the second minimum to connect
	//! @param s the State representating the saddle point
	//! @return whether or not the data structure has been changed
	virtual 
	bool 
	addSaddle(const size_t i, const size_t j, const State* const s) = 0;
	
	//! Sets a saddle point that connects two minima 
	//! @param m_i the first minimum to connect
	//! @param m_j the second minimum to connect
	//! @param s the State representating the saddle point
	//! @return whether or not the data structure has been changed
	virtual 
	bool 
	addSaddle(const State* const m_i, const State* const m_j, const State* const s) = 0;
	
	//! Adds an UNKNOWN minimum to the topology.
	//! NOTE : this operation should NOT change the indices of already existing
	//!        minima in the topology !
	//! FURTHER : the minimum is not checked if it is already known
	//! @param m the minimum to add
	//! @return the index of the minimum in the landscape
	virtual 
	const size_t 
	addMin(const State* const m) = 0;
	
	//! Access to the minimum index.
	//! @param m the minimum of that the index should be accessed
	//! @return  the index of the minimum, or INVALID_INDEX if 'm' is no known 
	//!          minimum
	virtual 
	const size_t 
	getMinIndex(const State* const m) const = 0;
	
	//! Access to the number of stored minima.
	//! @return the number of minima
	virtual 
	const size_t 
	getMinCount() const = 0;
	
	//! Access to a minimum.
	//! @param i index of the minimum to access
	//! @return the State with the specified minimum index
	virtual 
	const State* const 
	getMin(size_t i) const = 0;
	
	//! Access to the global minimum of the landscape with minimal free energy
	//! (mfe).
	//! @return the minimum with smallest energy (and smallest string 
	//!         representation for tiebreaking)
	virtual 
	const State* const 
	getMFEState() const = 0;

	//! Calculates the relative 'distance' between landscape topologies.
	//! On the LandscapeTopology level it is done on the minima only. For each
	//! minimum m present in only one of the two LTs a value of 
	//! MISSING_MINIMUM_PENALTY*(e^(-energy(m)/BOLTZMANN_KT)) is summed. The 
	//! resulting sum is normalized by the number of (unique) minima in both
	//! topologies.
	//! @param lt the landscape topology to calculate the distance to
	//! @return the distance
	virtual 
	double 
	getDistance(const LandscapeTopology* const lt) const;
	
	//! Sorts the minima ascending by their energy (mfe minimum gets index 0). 
	//! If energy is equal they are sorted by their string representation.
	virtual 
	void 
	sort() = 0;

	//! Check if the minima of the landscape are sorted ascendingly by their 
	//! energy (mfe minimum gets index 0). 
	//! If energy is equal they are sorted by their string representation.
	virtual 
	bool 
	isSorted() const = 0;

	//! Clears the data structure and removes all stored information to be 
	//! filled again afterwards. 
	virtual 
	void 
	clear() = 0;

	//! Writes the current landscape topology to stream.
	//! @param out the output stream to write to.
	//! @param writeGraph whether or not the output should contain a graph
	//!        representation of the topology in graphviz DOT format
	virtual 
	void
	write(	std::ostream & out, 
			const bool writeGraph = false ) const = 0;

	//! Fills the landscape topology from stream.
	//! @param in the input stream to read from.
	//! @param templateState the State template to create new State objects via
	//!        its 'fromString' function
	//! @return an error encoding and an error string :
	//!          0 = no error occured, read was successfull
	//!         -1 = landscape topology type not supported
	//!         -2 = state description differs from the given one
	//!         -3 = other read errors
	virtual 
	std::pair<int,std::string>
	read(	std::istream & in, 
			const State * const templateState ) = 0;

	
	
	///////////// STATIC IO MEMBERS ////////////////////////////////////////////
	
	
	
	//! Writes the descriptive header of a landscape topology to stream.
	//! @param out the output stream to write to.
	//! @param LT_ID the ID string of the topology to write
	//! @param minNumber the number of minima the topology holds, if minima are
	//!        grouped in the topology the number of groups has to be given.
	//! @param StateDescription a string description of the State model
	//!        represented by the topology
	static
	void
	writeHeader(	std::ostream & out, 
					const std::string& LT_ID,
					const size_t minNumber,
					const std::string & StateDescription );
	
	//! Writes a minimum of a landscape topology to stream.
	//! @param out the output stream to write to.
	//! @param minimum the minimum to add
	//! @param index the index of the minimum
	static
	void
	writeMinimum(	std::ostream & out, 
					const State* const minimum,
					const size_t index );

	//! Writes a group of minima of a landscape topology to stream. 
	//! NOTE: The FIRST minimum is the only one read by the read function and
	//! thus the representative for the whole group!
	//! @param out the output stream to write to.
	//! @param minima the minima to add
	//! @param index the index of the minimum
	static
	void
	writeMinimum(	std::ostream & out, 
					const std::vector<const State*> & minima,
					const size_t index );

	//! Writes a barrier of a landscape topology to stream.
	//! @param out the output stream to write to.
	//! @param barrier the barrier to add
	//! @param index the index of the barrier
	//! @param minConnected the minima connected by the barrier
	static
	void
	writeBarrier(	std::ostream & out, 
					const State* const barrier,
					const size_t index,
					const std::vector<size_t> & minConnected );
	
	
	//! Writes a saddle of a landscape topology to stream.
	//! @param out the output stream to write to.
	//! @param saddle the saddle to write
	//! @param index the index of the saddle
	//! @param minConnected the minima connected by the saddle
	static
	void
	writeSaddle(	std::ostream & out, 
					const State* const saddle,
					const size_t index,
					const std::vector<size_t> & minConnected );
	
	//! Parses the descriptive header of a landscape topology produced by 
	//! 'writeHeader' method.
	//! @param line the string that encodes the header information (encoding 
	//!        follows 'write' method)
	//! @param LT_ID the ID string of the topology in stream
	//! @param minNumber the number of minima the topology in stream holds, 
	//!        if minima are grouped in the topology the number of groups is 
	//!        given.
	//! @param StateDescription a string description of the State model
	//!        represented by the topology in the stream
	//! @return an error encoding and an error string :
	//!          0 = no error occured, read was successfull, the string contains
	//!              the remaining line content,
	//!         -3 = read error (see error string)
	static
	std::pair<int,std::string>
	parseHeader(	const std::string & line, 
					std::string& LT_ID,
					size_t& minNumber,
					std::string & StateDescription );

	//! Parses a minimum string encoding produced by 'writeMinimum' method and 
	//! adds the encoded minimum state to the given topology.
	//! Furthermore, it returns the remaining line content that can contain
	//! further minima that were merged with the first (already added) 'main'
	//! minimum. Thus a further parsing can be done later if necessary.
	//! @param toFill the barrier tree to add the barrier to
	//! @param line the string that encodes the minimum (encoding follows
	//!        'write' method)
	//! @param templateState the State template to create new State objects
	//!        via its 'fromString' function
	//! @param idx2min the mapping of minima indices in the string to the
	//!        indices in the given landscape topology 'toFill'
	//! @return an error encoding and an error string :
	//!          0 = no error occured, read was successfull, the string contains
	//!              the remaining line content,
	//!         -3 = read error (see error string)
	static
	std::pair<int,std::string>
	parseMinimum(	LandscapeTopology & toFill,
					const std::string & line, 
					const State * const templateState,
					Int2SizeT_MAP & idx2min );
	
	
	//! Parses a barrier string encoding produced by 'writeBarrier' method and 
	//! adds the encoded BARRIER as a SADDLE to the topology given.
	//! @param toFill the landscape topology to add the barrier to
	//! @param line the string that encodes the barrier (encoding follows
	//!        'write' method)
	//! @param templateState the State template to create new State objects
	//!        via its 'fromString' function
	//! @param idx2min the mapping of minima indices in the string to the
	//!        indices in the given landscape topology 'toFill'
	//! @return an error encoding and an error string :
	//!          0 = no error occured, read was successfull
	//!         -3 = read error (see error string)
	static
	std::pair<int,std::string>
	parseBarrier(	LandscapeTopology & toFill,
					const std::string & line, 
					const State * const templateState,
					const Int2SizeT_MAP & idx2min );

	
	//! Parses a saddle string encoding produced by 'writeBarrier' method and 
	//! adds the encoded direct saddle to the topology given.
	//! @param toFill the landscape topology to add the saddle to
	//! @param line the string that encodes the saddle (encoding follows
	//!        'write' method)
	//! @param templateState the State template to create new State objects
	//!        via its 'fromString' function
	//! @param idx2min the mapping of minima indices in the string to the
	//!        indices in the given landscape topology 'toFill'
	//! @return an error encoding and an error string :
	//!          0 = no error occured, read was successfull
	//!         -3 = read error (see error string)
	static
	std::pair<int,std::string>
	parseSaddle(	LandscapeTopology & toFill,
					const std::string & line, 
					const State * const templateState,
					const Int2SizeT_MAP & idx2min );

protected:  /////////////  HELPER FUNCTION  ////////////////////////////////////
	
	static
	double 
	getSortedMinimaDistance(	const std::vector<const State*> & m1,
								const std::vector<const State*> & m2 );
	
	
	  //! keyword used by the read and write method to identify information 
	  //! on the landscape topology type etc.
	static const std::string OUTPUT_KEY_ELL_HEAD;
	  //! keyword used by the read and write method to identify information 
	  //! on a minimum
	static const std::string OUTPUT_KEY_MINIMUM;
	  //! keyword used by the read and write method to identify information 
	  //! on a barrier
	static const std::string OUTPUT_KEY_BARRIER;
	  //! keyword used by the read and write method to identify information 
	  //! on a saddle point
	static const std::string OUTPUT_KEY_SADDLE;
	
	//! Writes a minima connection of a landscape topology to stream.
	//! @param out the output stream to write to.
	//! @param connectionState the State representing the connection
	//! @param index the index of the connection
	//! @param minConnected the minima connected 
	//! @param connectionKey the key used to encode the connection
	static
	void
	writeConnection(	std::ostream & out, 
						const State* const connectionState,
						const size_t index,
						const std::vector<size_t> & minConnected,
						const std::string& connectionKey  );
	
	//! Parses a string encoding of a minima connection of a topology
	//! produced by 'writeBarrier' method and returns the connection state and
	//! the list of minima connected.
	//! @param line the string that encodes the connection (encoding follows
	//!        'write' method)
	//! @param templateState the State template to create new State objects
	//!        via its 'fromString' function
	//! @param idx2min the mapping of minima indices in the string to the
	//!        indices in the given landscape topology 'toFill'
	//! @param connectionKey the key used to encode the connection
	//! @param connectedMinima OUTPUT parameter where the indices of the 
	//!        connected minima of the parsed connection are stored to
	//! @return a the parsed connection State and an error string if an error 
	//!         occured and the State is NULL
	static
	std::pair< State*, std::string>
	parseConnection(	const std::string & line, 
						const State * const templateState,
						const Int2SizeT_MAP & idx2min,
						const std::string& connectionKey,
						std::vector<size_t> & connectedMinima );
	
};

//##############################################################################
//##############################################################################



}// ell


#endif /*LANDSCAPETOPOLOGY_HH_*/
