#ifndef ELL_MINIMUMMAPPING_HH_
#define ELL_MINIMUMMAPPING_HH_


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


#include <ell/State.hh>
#include <ell/LandscapeTopology.hh>
#include <ell/Walk.hh>


namespace ell {

	/**
	 * Defines a mapping function of states to local minima of the landscape.
	 * 
	 * @author Martin Mann
	 */
	class MinimumMapping {
		
	public:
		
		virtual ~MinimumMapping() {}
		
		//! Gives the mapping of a State to a local minimum.
		//! @param toMap the State to map
		//! @return the local minimum toMap was projected to
		virtual
		State*
		getLocalMin( const State * const toMap ) const = 0;
		
		/*! Gives the index of the mapping of a State to a local minimum within
		 * a given LandscapeTopology.
		 * @param toMap the State to map
		 * @param lt the LandscapeTopology to get the index from
		 * @return the index of the local minimum toMap was projected to, or 
		 *         LandscapeTopology::INVALID_INDEX if not present in lt
		 */
		virtual
		size_t
		getLocalMinIndex(	const State * const toMap
							, const LandscapeTopology& lt ) const = 0;
		
	};
	
} // namespace ell


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


namespace ell {
	
	  /*!
	   * Maps a given ell::State via a gradient walk to a local minimum of the
	   * energy landscape.
	   *  
	   * @author Martin Mann
	   */
	class MM_GradientWalk : public MinimumMapping {
		
	public:
		//! Creates a gradient walk minimum mapping.
		MM_GradientWalk();
		//! destruction
		virtual ~MM_GradientWalk();
		
		//! Gives the local minimum that is reached via a gradient walk.
		//! @param toMap the State to map
		//! @return the local minimum toMap was projected to
		virtual
		State*
		getLocalMin( const State * const toMap ) const;
		
		/*! Gives the index of the local minimum reached via a gradient walk 
		 * within the given LandscapeTopology.
		 * @param toMap the State to map
		 * @param lt the LandscapeTopology to get the index from
		 * @return the index of the local minimum toMap was projected to, or 
		 *         LandscapeTopology::INVALID_INDEX if not present in lt
		 */
		virtual
		size_t
		getLocalMinIndex(	const State * const toMap
							, const LandscapeTopology& lt ) const;
		
	};

} // namespace ell



//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


namespace ell {
	
	  /*!
	   * Maps a given ell::State via a ell::GradientWalkHashed to a local 
	   * minimum of the energy landscape. This gradient walk implementation
	   * uses a hash to reduce computation.
	   * 
	   * @author Martin Mann
	   */
	class MM_GradientWalkHashed : public MinimumMapping {
		
	protected:
		
		  //! the gradient walk object in use
		ell::WalkGradientHashed gradWalk;

	public:
		//! Creates a gradient walk minimum mapping.
		MM_GradientWalkHashed( const size_t maxHashSize = 10000 );
		
		//! destruction
		virtual ~MM_GradientWalkHashed();
		
		//! Gives the local minimum that is reached via a gradient walk.
		//! @param toMap the State to map
		//! @return the local minimum toMap was projected to
		virtual
		State*
		getLocalMin( const State * const toMap ) const;
		
		/*! Gives the index of the local minimum reached via a gradient walk 
		 * within the given LandscapeTopology.
		 * @param toMap the State to map
		 * @param lt the LandscapeTopology to get the index from
		 * @return the index of the local minimum toMap was projected to, or 
		 *         LandscapeTopology::INVALID_INDEX if not present in lt
		 */
		virtual
		size_t
		getLocalMinIndex(	const State * const toMap
							, const LandscapeTopology& lt ) const;
		
		
		//! Sets the maximal number of elements allowed to store in the hash.
		//! @param s the maximal hash size
		virtual 
			void 
			setHashSize( const size_t s );

		//! Returns the maximal number of elements allowed to store in the hash.
		//! @return the maximal hash size
		virtual 
			const size_t
			getHashSize( void );
		
		//! Removes all elements currently stored in the hash.
		virtual 
			void
			clearHash( void );
	};

} // namespace ell


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#include <biu/LimitedHash.hh>

namespace ell {
	
	  /*!
	   * A MinimumMapping wrapper that internally uses a hash lookup and a 
	   * LandscapeTopology. Only if the state is not already within the hash 
	   * the wrapped mapping is used. Furthermore it is assumed, that all 
	   * minima are part of the LandscapeTopology.
	   *  
	   * @author Martin Mann
	   */
	class MM_Hashed_LT : public MinimumMapping {
		
	protected:
		
		//! the wrapped mapping 
		const MinimumMapping& wrappedMapping;
		
		//! the underlying LandscapeTopology that contains ALL minima we will
		//! map to.
		const LandscapeTopology& lt;
		
		typedef biu::LimitedHash< std::string, size_t > State2MinMap;

		typedef State2MinMap::value_type State2MinMapVal;
		//! the hash used to map a State in string representation to its 
		//! local minimum index in the LandscapeTopology
		mutable State2MinMap seenStates;
		
	public:
		
		/*! Creates a hashing wrapper.
		 * @param wrappedMapping the MinimumMapping used to map States to minima
		 * @param allMinima the LandscapeTopology that contains all minima the
		 * given wrappedMapping will map to. 
		 * @param maximal number of hashed states
		 */
		MM_Hashed_LT(	const MinimumMapping& wrappedMapping
						, const LandscapeTopology& allMinima
						, const size_t maxHashSize = (size_t)INT_MAX );
		
		//! destruction
		virtual ~MM_Hashed_LT();
		
		//! Gives the local minimum according to the wrapped MinimumMapping. 
		//! First it checks if the state was already checked before and is 
		//! stored in the internal hash. Otherwise it forwards the request 
		//! to the internally wrapped mapping.
		//! @param toMap the State to map
		//! @return the local minimum toMap was projected to
		virtual
		State*
		getLocalMin( const State * const toMap ) const;
		
		//! Gives the index of the local minimum in the given 
		//! LandscapeTopology according to the wrapped MinimumMapping. First
		//! it checks if the state was already checked before and is stored 
		//! in the internal hash. Otherwise it forwards the request to the
		//! internally wrapped mapping.
		//! NOTE : It is assumed that the given LandscapeTopology is the same  
		//! as the internally used one to enable the utilization of the 
		//! internal hash! Therefore it is ignored and the indices according to
		//! the internally used LandscapeTopology are given!
		//! @param toMap the State to map
		//! @param lt the LandscapeTopology to get the index from (ignored)
		//! @return the local minimum toMap was projected to
		virtual
		size_t
		getLocalMinIndex(	const State * const toMap
							, const LandscapeTopology& lt ) const;
		
	};

} // namespace ell


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#endif /*ELL_MINIMUMMAPPING_HH_*/
