#ifndef NEIGHBORGENERATOR_HH_
#define NEIGHBORGENERATOR_HH_


#include "ell/State.hh"

namespace ell {
	/*!
	 * This abstract class is used by all classes generating lists of neighbors.
	 * 
	 * @author Daniel Maticzka
	 */
	class NeighborGenerator
	{
	public:
	
		virtual ~NeighborGenerator(){};

		/*!
		 * Returns a NeighborListPtr to a list of neighbors to State s.
		 * @param s generate neighbors to this state
		 * @return autopointer to generated NeigborList
		 */
		virtual State::NeighborListPtr getNeighborList(const State* const s) const = 0;
	};
	
	/*!
	 * This subclass of NeighborGenerator generates a randomized list of
	 * Neighbors to a given State. Neighbour states are included multiple times.
	 * 
	 * @author Daniel Maticzka
	 */
	class NG_Random : public NeighborGenerator
	{
	protected:
		
		class NeighborList : public State::NeighborList
		{
		protected:
			//! the state of origin of the neighbors
			const State* const origin;
		public:
			NeighborList(const State* _origin);
			
			virtual ~NeighborList();

			/*! Returns a pointer to the first element of the
			 *  virtual list of neighbors of the state of origin.
			 *  Returns NULL if no neighbors exist.
			 * @param itstate constains state of the list iterator
			 * @return first element of NeighborList
			 */
			virtual State* first(State::NeighborList::ItState** itstate) const;

			/*! Returns a pointer to the next element of the
			 *  virtual list of neighbors of the state of origin.
			 *  Returns NULL if no more neighbors exist.
			 * @param itstate constains state of the list iterator
			 * @param elem this State is used for the next neighbor
			 * @return next element in NeighborList 
			 */
			virtual State* next(	State::NeighborList::ItState* itstate, 
									State* elem) const;
		};

	public:

		/*!
		 * Returns a NeighborListPtr to a randomized list of neighbors to State s.
		 * @param s generate neighbors to this state
		 * @return autopointer to generated NeigborList
		 */
		 State::NeighborListPtr getNeighborList(const State* const s) const;
	};
	
	/*!
	 * This subclass of NeighborGenerator generates a randomized list of
	 * Neighbors to a given State. Each neighbour state is included only once.
	 * 
	 * @author Daniel Maticzka
	 */
	class NG_SuccessiveRandom : public NeighborGenerator
	{
	public:

		/*!
		 * Returns a NeighborListPtr to a randomized list of neighbors to State s.
		 * @param s generate neighbors to this state
		 * @return autopointer to generated NeigborList
		 */
		 State::NeighborListPtr getNeighborList(const State* const s) const;
	};

	/*!
	 * This subclass of NeighborGenerator generates a list of
	 * Neighbors to a given State.
	 * 
	 * @author Daniel Maticzka
	 */
	class NG_Successive : public NeighborGenerator
	{
	public:

		/*!
		 * Returns a NeighborListPtr to a list of neighbors to State s.
		 * @param s generate neighbors to this state
		 * @return autopointer to generated NeigborList
		 */
		 State::NeighborListPtr getNeighborList(const State* const s) const;
	};
	
} // namespace ell

#endif /*NEIGHBORGENERATOR_HH_*/
