#ifndef RATEMATRIX_HH_
#define RATEMATRIX_HH_

#include <biu/MatrixSparse.hh>

#include <ell/MinimumMapping.hh>

namespace ell {

	  /*! Definition of a rate matrix. It is not ensuring that it is a square
	   * matrix! Has to be done by the user!
	   * 
	   * @author Martin Mann - http://www.bioinf.uni-freiburg.de/~mmann/
	   * 
	   */
	class RateMatrix : public biu::MatrixSparseC<double> {
		
	public:
		
			//! creates an empty rate matrix of dimensions (0,0)
			//! @param defVal the default value to use for new entries
		explicit 
		RateMatrix(const double defVal);
			//! creates a matrix of given dimension
			//! @param r the row number of the matrix
			//! @param c the column number of the matrix
			//! @param val the default value for new matrix entries
		RateMatrix(const size_t r, const size_t c, const double val);
			//! copy constructor including matrix resizing
			//! @param mat the matrix this matrix should be a copy of
		RateMatrix(const RateMatrix &mat);
		
	public:

		  /*! Calculates columnwise an RMSD of the corresponding columns, i.e.
		   *  an RMSD of the transition rates for each minimum
		   *  ASSUMPTION: R1 and R2 describe the same set of states, such that
		   *  R1[to][from] describes the same transition than R2[to][from]. So 
		   *  the dimension of R1 and R2 have to be the same!
		   *  @param R1 the first rate matrix
		   *  @param R2 the second rate matrix
		   *  @return a column RMSD of the transitions for each state covered 
		   *          by R1 and R2
		   */
		static
		std::vector<double>
		RMSD( const RateMatrix & R1, const RateMatrix & R2);
			
		  /*! Calculates columnwise the Kullback-Leibler divergence of the 
		   *  corresponding columns, i.e. the information divergence of the 
		   *  transition rates for each minimum which are probability 
		   * distributions.
		   *  
		   *  FORMULA : for each column c : sum_r (R2[r,c] * log(R2[r,c]/R1[r,c])) 
		   *  
		   *  ASSUMPTION: R1 and R2 describe the same set of states, such that
		   *  R1[to][from] describes the same transition than R2[to][from]. So 
		   *  the dimension of R1 and R2 have to be the same!
		   *  
		   *  @param R1 the first rate matrix
		   *  @param R2 the second rate matrix
		   *  @return the Kullback-Leibler divergence for each column of 
		   *          transitions for each state covered by R1 and R2
		   */
		static
		std::vector<double>
		KLD( const RateMatrix & R1, const RateMatrix & R2);
	
		  /*! Normalizes the given rate matrix to gain a stochastic matrix,
		   * i.e. each column sum equals 1.0.
		   * 
		   * @param R the matrix to normalize
		   */
		static
		void
		normalize( RateMatrix & R );
	
	};
	

} // namespace ell

#endif /*RATEMATRIX_HH_*/
