#ifndef ELL_RATESAMPLING_BASIN_HH_
#define ELL_RATESAMPLING_BASIN_HH_

#include <biu/HashMap.hh>
// include best available hash or map
#if HAVE_UNORDERED_MAP == 1
	#include <unordered_map>
#elif HAVE_TR1_UNORDERED_MAP == 1
	#include <tr1/unordered_map>
#elif HAVE_GNU_HASH_MAP == 1
	#include <ext/hash_map>
#else
	#include <map>
#endif

#include "ell/LandscapeTopology.hh"
#include "ell/MinimumMapping.hh"

#include "ell/RateMatrix.hh"

#include <cstdlib>

namespace ell
{

	 /*! Implements a sampling procedure that estimates transition probabilities
	  * of a given macro-state in the energy landscape to all directly adjacent
	  * macro-states. The sampling is based on a Metropolis Monte-Carlo
	  * sampling utilizing a given relative temperature kT.
	  * 
	  * For details on the method refer to :
	  * 
	  * "Efficient exploration of energy landscapes", 
	  * Martin Mann and Konstantin Klemm,
	  * 2009 (submitted)
	  * 
	  * 
	  * @author Martin Mann - http://www.bioinf.uni-freiburg.de/~mmann/
	  */
	class RateSampling_Basin
	{
	public:
		
		  //! rate map to use, i.e. one column within the transition matrix
		typedef ell::RateMatrix::EntryMap RateMap;			
		
		
		  /*! Helper superclass that is made to track the sampling progress
		   * along a single run.
		   */
		class RateSamplingReporter {
		public:
			
			  /*! Reports a transition matrix column of the given local minimum
			   * after that the given number of steps has been performed.
			   * @param R the transition probabilities
			   * @param curMinID the index of the corresponding local minimum
			   *         in the underlying LandscapeTopology
			   * @param doneSteps the performed sampling steps to obtain the
			   *         given R 
			   */
			virtual
			void
			report(	const RateMap& R
					, const size_t curMinID
					, const size_t doneSteps
					) = 0;
		};
		
		  /*! Data class to store information on the sampling performed.
		   */
		class SamplingStatistics {
		public:
			  //! number of sampling steps performed
			size_t iterations;
			  //! virtual time consumption calculated based on the transition
			  //! probabilities of the performed transitions
			double simTime;
			  /*! Construction
			   * @param iter the number of sampling steps performed
			   * @param time the virtual time consumed
			   */
			SamplingStatistics( const size_t iter, const double time );
		};
		
		
	protected:
		
		 /*! Utility class to store neighbor transition information for a given
		  * neighbored state to enable a fast lookup of possible transitions 
		  * instead of a recalculation.
		  */
		class NeighData {
		public:
			  //! the string representation of the neighbor
			std::string ID;
			  //! the minimum index of the neighbor within the underlying
			  //! LandscapeTopology
			size_t minIdx;
			  //! the transition probability to this neighbor
			double prob;
			  //! the energy of this state
			double energy;
			
			  //! Default construction.
			NeighData();
			
			  /*! Construction
			   * @param ID_ the string representation of the neighbor
			   * @param minIdx_ the minimum index of the neighbor within the 
			   *  underlying LandscapeTopology
			   * @param prob_ the transition probability to this neighbor
			   * @param energy_ the energy of this state
			   */
			NeighData(	const std::string& ID_
						, const size_t minIdx_
						, const double prob_
						, const double energy_);
		};
		
		  /*! Specialized sorting class to order the neighbored state 
		   * information as needed for the algorithm.
		   */
		class NeighDataSorter {
			  //! the local minimum index of interest to which all other
			  //! minima indices are set in relation to
			const size_t curMinIdx;
			
		public:
			  /*! Construction
			   *  @param curMinIdx_ the local minimum index of interest to
			   *    which all other minima indices are set in relation to
			   */
			NeighDataSorter(const size_t curMinIdx_);
			
			  /*! Comparison operator that returns a dedicated strict less
			   * ordering (i<j) on neighboring data objects.
			   * @param i the first object to compare
			   * @param j the second object to compare
			   * @return returns (|i.minIdx-curMinIdx| < |j.minIdx-curMinIdx|)
			   *      or (equal and (i.prob > j.prob))
			   *      or (both equal and (i.ID < j.ID))
			   */ 
			bool operator() (	const NeighData &i, const NeighData &j );
		};

		// set typedef for best available hash or map
		#if HAVE_UNORDERED_MAP == 1
			typedef std::unordered_map< std::string, std::vector<NeighData> > NeighDataMap;
		#elif HAVE_TR1_UNORDERED_MAP == 1
			typedef std::tr1::unordered_map< std::string, std::vector<NeighData> > NeighDataMap;
		#elif HAVE_GNU_HASH_MAP == 1
			typedef __gnu_cxx::hash_map< std::string, std::vector<NeighData>, hash_string > NeighDataMap;
		#else
			typedef std::map< std::string, std::vector<NeighData> > NeighDataMap;
		#endif
		
			
		 /*! Holds the neighbor information of the current basin sampled for
		  * faster look-up based neighbor access. 
		  */
		NeighDataMap neighData;
			
	public:
		
		  /*! the maximal number of minima to sample the rates for, i.e. the 
		   * maximal dimension of the rate matrix to fill and return.
		   */
		size_t maxMinima;
		
		  /*! base of the number of steps after that an output occures. after 
		   * each (outsteps^i) steps an output is triggered to the 
		   * rateSamplingReporter
		   */
		size_t outSteps;
		
		  /*! the rate sampling reporter to forward the output to if non-NULL.
		   */
		RateSamplingReporter* rateSamplingReporter;
		
		  /*! the relative temperature used to calculate the Boltzmann weights
		   * of states according to exp(-E/kT).
		   */
		double kT;
		
		  /*! Maximal number of simulation steps to wait until to proceed to 
		   * the next counted rate sampling step. This is needed to avoid very
		   * long stays in deep local minima.
		   */ 
		size_t maxWaitingSteps;
		
		  /*! the maximal neighbor number of any state within the current 
		   * energy landscape. This is needed to calculate correct probabilities
		   * for landscapes where states have different numbers of neighbors.
		   * Set to zero if the number of neighbors of the current state has
		   * to be used.
		   */
		size_t maxNeighNumber;
		
	public:
		
		  //! Construction that sets all default parameters.
		RateSampling_Basin();
		
		  //! Destruction.
		virtual ~RateSampling_Basin();
		
		
		  /*!
		   * Rate sampling per basin. Performs a rate sampling for each 
		   * basin/minimum of the energy landscape given. If new adjacent 
		   * basins are found their minima are added to the landscape topology
		   * and their basins are sampled as well.
		   * 
		   * NOTE: Only the first maxMinima basins of the landscape topology are
		   * considered for rate calculation. The rates to all other minima are
		   * ignored.
		   * 
		   * NOTE: This might lead to a number of minima higher than maxMinima.
		   * The corresponding minima are added to the landscape topology but 
		   * the transitions to their basins are ignored for the rate matrix
		   * calculation, i.e. they are assumed to be non-existing.
		   * 
		   * @param lt the landscape topology that contains the initially known
		   *           minima of interest and that is extended during the rate
		   *           sampling
		   * @param minMap the minimum mapping to use to derive the according
		   *           minimum for each state
		   * @param rates the rate matrix to fill along the sampling
		   * @param steps2sample number of steps to perform within each basin
		   *           to sample the according rates
		   * 
		   * @return statistics on the sampling performed
		   */
		SamplingStatistics
		sampleRates(
				LandscapeTopology & lt,
				const MinimumMapping & minMap,
				RateMatrix & rates,
				const size_t steps2sample
				);
		
		  /*!
		   * Performs the rate sampling for a given basin. If new adjacent 
		   * basins are found their minima are added to the landscape topology.
		   * 
		   * NOTE: This might lead to a number of minima higher than maxMinima.
		   * The corresponding minima are added to the landscape topology but 
		   * the transitions to their basins are ignored for the rate matrix
		   * calculation, i.e. they are assumed to be non-existing.
		   * 
		   * @param lt the landscape topology that contains the initially known
		   *           minima of interest and that is extended during the rate
		   *           sampling
		   * @param curMin the minimum index of the current basin to sample 
		   * @param minMap the minimum mapping to use to derive the according
		   *           minimum for each state
		   * @param rates the container for the rates to the adjacent minima
		   * @param steps2sample number of steps to perform within each basin
		   *           to sample the according rates
		   * 
		   * @return statistics on the sampling performed
		   */
		SamplingStatistics
		sampleRates(
				LandscapeTopology & lt,
				const size_t curMin,
				const MinimumMapping & minMap,
				RateMap & rates,
				const size_t steps2sample
				);
	};

} // namespace ell

#endif /*ELL_RATESAMPLING_BASIN_HH_*/
