#ifndef STATECOLLECTOR_HH_
#define STATECOLLECTOR_HH_

#include "ell/State.hh"

//##############################################################################
//##############################################################################

namespace ell 
{
	/*!
	 * This is an abstract interface used by all state collecting or processing
	 * classes (SC_*).
	 * 
	 * @author Daniel Maticzka
	 * @author Martin Mann
	 */
	class StateCollector
	{
	public:	
	
		virtual
		~StateCollector() {}
		
		  //! This function makes the collector manage the given State.
		  //! @param s the added State
		virtual void add(const State& s) = 0;
		
		  //! Returns number of added States.
		  //! @return the number of added states
		virtual size_t size() const = 0;
		
		  //! Returns last added State.
		  //! @return last added State
		virtual const State* const getLastAdded() const = 0;
		
	};
	
} // namespace ell


//##############################################################################
//##############################################################################

namespace ell
{
	/*!
	 * This class simply keeps track of the number of added States. 
	 * 
	 * @author Daniel Maticzka
	 */
	class SC_Counting : public StateCollector
	{
	protected:
	
		  //! the last added State
		State* lastAdded;
		  //! the number of added States
		size_t stateCount;
		
	public:	
	
		SC_Counting();
		
		virtual
		~SC_Counting();
	
		  //! Increments State counter.
		  //! @param s the added State
		virtual void add(const State& s);
		
		  //! Returns number of added States.
		  //! @return the number of added states
		virtual size_t size() const;
		 
		  //! Returns last added State.
		  //! @return last added State
		virtual const State* const getLastAdded() const;
		
	};
	
} // namespace ell


//##############################################################################
//##############################################################################

#include <list>

namespace ell 
{
	
	/*!
	 * This subclass of of StateCollector keeps a list of all added States.
	 * 
	 * @author Daniel Maticzka
	 */
	class SC_Listing : public StateCollector
	{
	public:
	
		typedef std::list<State*> List;
		 
	protected:
	
		  //! the internal storage of all added states 
		List stateList;
		
	public:
	
		SC_Listing();
		~SC_Listing();

		  //! Adds State to the state list.
		  //! @param s the added State
		virtual void add(const State& s);
		
		  //! Returns number of added States.
		  //! @return the number of added states
		virtual size_t size() const;
		 
		  //! Returns last added State.
		  //! @return the last added State
		virtual const State* const getLastAdded() const;
		
		///////////  LIST HANDLING  ////////////////////////
	
		  //! Returns a list of all accepted States.
		  //! @return the State list
		virtual const List& getList() const;
		
	};
	
} // namespace ell


//##############################################################################
//##############################################################################

#include <vector>

namespace ell 
{
	
	/*!
	 * This subclass of of StateCollector keeps a list of the k last added 
	 * States.
	 * 
	 * @author Martin Mann
	 */
	class SC_ListingK : public StateCollector
	{
	public:
	
		typedef std::vector<State*> List;
		 
	protected:
	
		  //! the internal storage of all added states 
		List stateList;
		  //! index of the last added state in stateList
		size_t curEnd;
		
		  //! total number of added States
		size_t addedTotal;
		
	public:
	
		  //! @param length the length of the list to store, i.e. the number of
		  //!        last added states.
		SC_ListingK(const size_t length);
		~SC_ListingK();

		  //! Adds State to the state list.
		  //! @param s the added State
		virtual void add(const State& s);
		
		  //! Returns number of added States.
		  //! @return the number of added states
		virtual size_t size() const;
		 
		  //! Returns last added State.
		  //! @return the last added State
		virtual const State* const getLastAdded() const;
		
		///////////  LIST HANDLING  ////////////////////////
		
		 /*!
		  * Iterator for the access of the stored States. 
		  */
		class iterator {
		protected:
			const size_t start;
			size_t i;
			const SC_ListingK* const parent;
		public: 
			
			iterator( const size_t start_, const SC_ListingK* const parent_)
				: start(start_), i(start_), parent(parent_)
			{
			}
			
			const State* const operator*() const {
				return parent->stateList[i];
			}
			
			bool operator++() {
				i = (i+1) % parent->stateList.size();
				return i != start;
			}
			
			bool operator++( int ) {
				i = (i+1) % parent->stateList.size();
				return i != start;
			}
		};

		virtual iterator getList() const;
		
	};
	
} // namespace ell


//##############################################################################
//##############################################################################

#include <list>

namespace ell 
{
	
	/*!
	 * This subclass of StateCollector keeps a list of all added States and
	 * stores the compressed string representation of each State.
	 * 
	 * @author Daniel Maticzka
	 * @author Martin Mann
	 */
	class SC_ListingCompr : public SC_Counting
	{
	public:
	
		typedef std::list<CSequence> List;
		 
	protected:
	
		  //! the internal storage of all added states 
		List stateList;

	public:
	
		SC_ListingCompr();
		~SC_ListingCompr();

		  //! This function stores the compressed string representation  
		  //! in a list.
		  //! @param s the added State
		virtual void add(const State& s);
		
		///////////  LIST HANDLING  ////////////////////////
	
		  //! Returns the list of compressed string representations of all 
		  //! managed States.
		  //! @return the compressed sequence list
		virtual const List& getList() const;

	};
	
} // namespace ell

//##############################################################################
//##############################################################################

namespace ell
{

	  /*! A StateCollector that stores the State which is maximal according
	   * to the State::less operator.
	   * 
	   * @author Martin Mann, Hannes Kochniss
	   */
	class SC_Maximum : public SC_Counting
	{
	protected:
		State* maxEState;
		
	public:
		SC_Maximum();
		virtual ~SC_Maximum();
		
		  //! Checks if the new State is of higher order than the current 
		  //! maxEState and forwards the call to the super class.
		  //! @param s the added State
		virtual
		void
		add(const State& s);
		
		  //! Returns the maximal State seen so far.
		  //! @return the State 
		const State* const
		getMaximumState() const;
	};

} // namespace ell

//##############################################################################
//##############################################################################

#include <iostream>
#include <string>

namespace ell
{

	/*! A StateCollector that prints each reported state to the given outstream
	 * using the additional given flag to separate the printed States.
	 * 
	 * @author Martin Mann 
	 */
	class SC_Outstream : public SC_Counting
	{
	protected:
		  //! the stream to write the states to
		std::ostream& out;
		
		  //! additional string written to stream AFTER each state
		const std::string strAdded;
		
	public:
	
		  //! Constructs a StateCollector that writes all States to out stream.
		  //! @param out the outstream to write to
		  //! @param toAppend	string written to stream AFTER each state
		SC_Outstream(	std::ostream& out 
						, const std::string& toAppend);
		
		virtual ~SC_Outstream();
	
		  //! This function is used to track all accepted intermediate States.
		  //! @param s the accepted State
		virtual void add(const State& s);
			
	};

} // namespace ell

//##############################################################################
//##############################################################################

#include <ell/LandscapeTopology.hh>

namespace ell
{

	  /*! This StateCollector implementation counts the number of collected
	   * States and calculates the canonical partition function Z of these 
	   * States. Z = sum( e^(-E(State)/kT) ) is the sum of the Boltzmann weights
	   * of all States. The kT value is derived from 
	   * LandscapeTopology::BOLTZMANN_KT.
	   * 
	   * NOTE : if the energy of collected states is << 0 one will face
	   * immense rounding errors due to the summing of e^(-energy). In this case
	   * we recommend the SC_PartitionFunction_Shifted state collector which 
	   * uses a more advanced handling of the Boltzmann weights based on energy
	   * differences instead of pure energies (without negative impact in accuracy).
	   * 
	   */
	class SC_PartitionFunction : public SC_Counting
	{
	public:
		  //! Construction
		SC_PartitionFunction();
		  //! Destruction
		virtual ~SC_PartitionFunction();
		
		  //! This function makes the collector manage the given State.
		  //! @param s the added State
		virtual void add(const State& s);

	protected:
		
		  //! the current canonical partition function of the collected States
		double Z;
		
	public:
		
		  /*! Access to the canonical partition function of the collected 
		   * States.
		   * @return the partition function Z
		   */ 
		double getZ( void ) const;
		
		  /*! Calculates the free energy of the collected State ensemble based
		   * on the canonical partition function : E = -kT * ln(Z) 
		   * @return the energy of the State ensemble
		   */ 
		double getEnergy( void ) const;
	};
	

//##############################################################################
//##############################################################################

	
	  /*! This StateCollector implementation counts the number of collected
	   * States and calculates the canonical partition function Z of these 
	   * States. Z = sum( exp(-E(State)/kT) ) is the sum of the Boltzmann weights
	   * of all States. The kT value is derived from 
	   * LandscapeTopology::BOLTZMANN_KT.
	   * 
	   * NOTE: This class is almost always a better choice comp. to SC_PartitionFunction, 
	   * as any energies < -1 can incure severe rounding errors and even overflow, esp. when 
	   * summing up many states of low energy
	   * Here we use the energy difference to the minimal energy of the 
	   * landscape the States are from and sum over shifted Boltzmann weights.
	   * The final Z and energy are calculated from that. 
	   * 
	   * The shifted Z calculation is based on the minimal energy of the
	   * Landscape MinE. Thus we sum 
	   * Z' = sum( exp(-(E(State)-MinE)/kT) ) = sum( exp((MinE-E(State))/kT) ).
	   * 
	   * The final Z is than calculated using Z = Z' * exp(-MinE/kT).
	   * 
	   */
	class SC_PartitionFunction_Shifted : public SC_Counting
	{
	public:
		
		  //! Construction
		  //! @param lt The landscape topology from that the minimum free 
		  //! energy is used to make the energy shift of the Boltzmann weights.
		SC_PartitionFunction_Shifted( const LandscapeTopology& lt );
		
		  //! Construction
		  //! NOTE : If this constructor is used YOU have to ensure that the  
		  //! energies of all collected States are greater equal >= to this 
		  //! value !
		  //! @param MinE the minimal energy to use. NOTE: has to be >= to
		  //!        the energies of all collected States.
		  //! energy is used to make the energy shift of the Boltzmann weights.
		SC_PartitionFunction_Shifted( const double MinE );
		
		  //! Destruction
		virtual ~SC_PartitionFunction_Shifted();
		
		  //! This function makes the collector manage the given State.
		  //! @param s the added State
		virtual void add(const State& s);

	protected:
		
		  //! The current canonical partition function of the collected States
		  //! shifted by the minimum free energy of the landscape.
		double Z_shifted;
		
		  //! Minimal energy of the landscape
		const double MIN_E;
		
	public:
		
		  /*! Access to the canonical partition function of the collected 
		   * States. Z = Z' * exp(-MinE/kT)
		   * @return the partition function Z
		   */ 
		double getZ( void ) const;
		
		  /*! Calculates the free energy of the collected State ensemble based
		   * on the canonical partition function : E = MIN_E -(kT * ln(Z')) 
		   * @return the energy of the State ensemble
		   */ 
		double getEnergy( void ) const;
	};

} // namespace ell

//##############################################################################
//##############################################################################

#endif /*STATECOLLECTOR_HH_*/
